﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using HeuristicLab.Persistence.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Persistence {
  [Transformer("78556C88-0FEE-4602-95C7-A469B2DDB468", 100)]
  internal sealed class StorableClassTransformer : TransformerBase {
    public override bool CanTransformType(Type type) {
      return StorableClassAttribute.IsStorableClass(type);
    }
    public override PersistenceData ToData(object o, PersistenceMapper mapper) {
      var type = o.GetType();
      var typeInfo = mapper.GetTypeInfo(type);
      var attribute = typeInfo.StorableClassAttribute;
      var data = new CompositeData(Id);
      mapper.Cache(o, data);
      data.TypeId = mapper.GetTypeId(type);
      // data.GuidId = attribute.Guid;
      // data.Version = attribute.Version;

      var emptyArgs = new object[0];
      foreach (var hook in typeInfo.BeforeSerializationHooks) {
        hook.Invoke(o, emptyArgs);
      }
      var components = new Dictionary<uint, uint>();
      foreach (var componentInfo in typeInfo.Fields) {
        var field = (FieldInfo)componentInfo.MemberInfo;
        var component = mapper.GetDataId(field.GetValue(o));
        components.Add(mapper.GetStringId(componentInfo.Name), component);
      }
      foreach (var componentInfo in typeInfo.Properties.Where(x => x.Readable)) {
        var property = (PropertyInfo)componentInfo.MemberInfo;
        var component = mapper.GetDataId(property.GetValue(o, null));
        components.Add(mapper.GetStringId(componentInfo.Name), component);
      }
      data.ComponentNameIds = components.Keys.ToArray();
      data.ComponentIds = components.Values.ToArray();
      return data;
    }

    public override object ToObject(PersistenceData data, PersistenceMapper mapper) {
      var composite = (CompositeData)data;
      var type = mapper.GetType(composite.TypeId);
      var typeInfo = mapper.GetTypeInfo(type);
      var o = mapper.CreateInstance(type);
      mapper.Cache(data, o);

      var components = new Dictionary<uint, uint>();
      for (int i = 0; i < composite.ComponentNameIds.Length; i++) {
        components.Add(composite.ComponentNameIds[i], composite.ComponentIds[i]);
      }

      foreach (var componentInfo in typeInfo.Fields) {
        var field = (FieldInfo)componentInfo.MemberInfo;
        uint componentId;
        components.TryGetValue(mapper.GetStringId(componentInfo.Name), out componentId);
        field.SetValue(o, componentId != 0 ? mapper.GetObject(componentId) : componentInfo.StorableAttribute.DefaultValue);
      }
      foreach (var componentInfo in typeInfo.Properties.Where(x => x.Writeable)) {
        var property = (PropertyInfo)componentInfo.MemberInfo;
        uint componentId;
        components.TryGetValue(mapper.GetStringId(componentInfo.Name), out componentId);
        property.SetValue(o, componentId != 0 ? mapper.GetObject(componentId) : componentInfo.StorableAttribute.DefaultValue, null);
      }
      var emptyArgs = new object[0];
      foreach (var hook in typeInfo.AfterDeserializationHooks) {
        hook.Invoke(o, emptyArgs);
      }
      return o;
    }
  }
}
